#BOTSY Art Files Generator By BOTSY LLC (C) 2022
#Version 3.0
#Date Release: August 26, 2023 8:04pm
#Website: www.BOTSY.com
#Contact Info: Liza Kholodkova - liza@botsy.com
#Based on J TECH PHOTONICS LASER TOOL V2.4 4/20/2022
#Special thanks to Pedro Arroyo

import os       #module for working with folders of the system

#varaibles moved from the inx file here to remove them from having user to input.
passes = 1
pass_depth = 0.5
invert_y_axis = False
horizontal_offset = 0
vertical_offset = 0
scaling_factor = 1
tool_power_command = "G01 Z-1.0;"
tool_off_command = "G00 Z30.0;"  #has to be high value
tool_off_command2 = "G00 Z15.0;" #has to be lower then tool_off_command!
dwell_time = 0
header_path = None
footer_path = None
do_z_axis_start = False
z_axis_start = 0.1
do_laser_off_start = False
do_laser_off_end = True
unit = "mm"


from lxml import etree      #module for inporting from inx and xml files
from xml.etree import ElementTree as xml_tree
from inkex import EffectExtension, Boolean

from svg_to_gcode.svg_parser import parse_root, Transformation, debug_methods
from svg_to_gcode.geometry import LineSegmentChain
from svg_to_gcode.compiler import Compiler, interfaces
from svg_to_gcode import TOLERANCES

#Pedro BEGIN Supports warning messages
from tkinter import messagebox
#Pedro END

svg_name_space = "http://www.w3.org/2000/svg"
inkscape_name_space = "http://www.inkscape.org/namespaces/inkscape"
sodipodi_name_space = "http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"

inx_filename = "BOTSYv3_0.inx"

def generate_custom_interface(laser_off_command2, laser_off_command, laser_power_command):
    """Wrapper function for generating a Gcode interface with a custom laser power command"""

    class CustomInterface(interfaces.Gcode):
        """A Gcode interface with a custom laser power command"""

        def __init__(self):
            super().__init__()

        def laser_off(self):
            return f"{laser_off_command}"
#Liza Edit: new varibale for footer custom heading: tool_off_command2 = G00 Z30.0; new variable is in .inx file
        def laser_off2(self):
            return f"{laser_off_command2}"
#Liza Edit
        def set_laser_power(self, _):
            return f"{laser_power_command}"

    return CustomInterface


class GcodeExtension(EffectExtension):
    """Inkscape Effect Extension."""

    def __init__(self):
        EffectExtension.__init__(self)

#Pedro BEGIN Sends a warning message to the user
    def warning(self, msg):
        messagebox.showwarning(title="BOTSY Warning", message=msg)
#Pedro END

    def effect(self):
        """Takes the SVG from Inkscape, generates gcode, returns the SVG after adding debug lines."""

        root = self.document.getroot()

# Pedro BEGIN - Checks preconditions and warns user if they are violated
        # TODO Check to see if the svg is outside the canvas and warn the user
        #self.warning("Drawing out of bounds.")
#Pedro END


        # Change svg_to_gcode's approximation tolerance
        TOLERANCES["approximation"] = float(self.options.approximation_tolerance.replace(',', '.'))

        # Construct output path
        # Construct output path
        if self.options.filename_dynamic and self.document_path():
            filename, extension = self.document_path().split('.')
            filename = filename.split('/')[-1] + '.gcode'
            output_path = os.path.join(self.options.directory, filename)
        else:
            output_path = os.path.join(self.options.directory, self.options.filename)

        if self.options.filename_suffix:
            try:
                filename, extension = output_path.split('.')
            except:
                self.msg("Error in file folder! Check that the folder exists.")
                exit(1)

            n = 1
            while os.path.isfile(output_path):
                output_path = filename + str(n) + '.' + extension
                n += 1

        # Load header and footer files
        header = []
        if header_path is not None:
            if os.path.isfile(header_path):
                with open(header_path, 'r') as header_file:
                    header = header_file.read().splitlines()
            elif header_path != os.getcwd():  # The Inkscape file selector defaults to the working directory
                self.debug(f"Header file does not exist at {header_path}")

        footer = []
        if footer_path is not None:
            if os.path.isfile(footer_path):
                with open(footer_path, 'r') as footer_file:
                    footer = footer_file.read().splitlines()
            elif footer_path != os.getcwd():
                self.debug(f"Footer file does not exist at {footer_path}")

        # Customize header/footer Liza edit added 3rd variable: self.options.tool_off_command2,
        custom_interface = generate_custom_interface(tool_off_command2, tool_off_command, tool_power_command)
        interface_instance = custom_interface()

        if do_laser_off_start:
            header.append(interface_instance.laser_off())
        if do_laser_off_end:
            footer.append(interface_instance.laser_off2()) #liza edit, change laser_off to laser_off2

        header.append(interface_instance.set_movement_speed(self.options.travel_speed))
        if do_z_axis_start:
            header.append(interface_instance.linear_move(z=z_axis_start))
        if self.options.move_to_origin_end:
            footer.append(interface_instance.linear_move(x=0, y=0))

        # Generate gcode
        gcode_compiler = Compiler(custom_interface, self.options.travel_speed, self.options.cutting_speed,
                                  pass_depth, dwell_time=dwell_time, custom_header=header,
                                  custom_footer=footer, unit=unit)

        transformation = Transformation()

        transformation.add_translation(horizontal_offset, vertical_offset)

#Pedro BEGIN Adds transformation to support anchoring the botsy workspace to the bottom of the document
        transformation.add_translation(0, -self.svg.viewbox_height + self.options.bed_height)
#Pedro END

        transformation.add_scale(scaling_factor)

        if self.options.machine_origin == "center":
            transformation.add_translation(-self.options.bed_width / 2, self.options.bed_height / 2)
        elif self.options.machine_origin == "top-left":
            transformation.add_translation(0, self.options.bed_height)

        curves = parse_root(root, transform_origin=not invert_y_axis, root_transformation=transformation,
                            canvas_height=self.options.bed_height)

        gcode_compiler.append_curves(curves)
        gcode_compiler.compile_to_file(output_path, passes=passes)

        # Draw debug lines
        self.clear_debug()
        if self.options.draw_debug:
            self.draw_debug_traces(curves)
            self.draw_unit_reference()
            self.select_non_debug_layer()

        return self.document

    def draw_debug_traces(self, curves):
        """Traces arrows over all parsed paths"""

        root = self.document.getroot()
        origin = self.options.machine_origin
        bed_width = self.options.bed_width
        bed_height = self.options.bed_height

        group = etree.Element("{%s}g" % svg_name_space)
        group.set("id", "debug_traces")
        group.set("{%s}groupmode" % inkscape_name_space, "layer")
        group.set("{%s}label" % inkscape_name_space, "debug traces")

        group.append(
            etree.fromstring(xml_tree.tostring(debug_methods.arrow_defs(arrow_scale=self.options.debug_arrow_scale))))

        for curve in curves:
            approximation = LineSegmentChain.line_segment_approximation(curve)

            change_origin = Transformation()

#Pedro BEGIN Adds transformation to support anchoring the botsy workspace to the bottom of the document
            change_origin.add_translation(0, self.svg.viewbox_height - self.options.bed_height)
#Pedro END

            if not invert_y_axis:
                change_origin.add_scale(1, -1)
                change_origin.add_translation(0, -bed_height)

            if origin == "center":
                change_origin.add_translation(bed_width / 2, bed_height / 2)



            path_string = xml_tree.tostring(
                debug_methods.to_svg_path(approximation, color="red", opacity="0.5", stroke_width=f"{self.options.debug_line_width}px",
                                          transformation=change_origin, draw_arrows=True)
            )

            group.append(etree.fromstring(path_string))

        root.append(group)

    def draw_unit_reference(self):
        """Draws reference points to mark the bed's four corners"""
        root = self.document.getroot()
        origin = self.options.machine_origin
        bed_width = self.options.bed_width
        bed_height = self.options.bed_height

        group = etree.Element("{%s}g" % svg_name_space)
        group.set("id", "debug_references")
        group.set("{%s}groupmode" % inkscape_name_space, "layer")
        group.set("{%s}label" % inkscape_name_space, "debug reference points")

#Pedro BEGIN Changes to support anchoring the bosty workspace to the bottom of the document
        #reference_points_svg = [(0, 0), (0, bed_height), (bed_width, 0), (bed_width, bed_height)]
        reference_points_svg = {
            "top-left": (0, self.svg.viewbox_height - bed_height),
            "bottom-left": (0, self.svg.viewbox_height),
            "top-right": (bed_width, self.svg.viewbox_height - bed_height),
            "bottom-right": (bed_width, self.svg.viewbox_height)
        }

        reference_points_gcode = {
            "bottom-left": [(0, bed_height), (0, 0), (bed_width, bed_height), (bed_width, 0)],
            "top-left": [(0, 0), (0, bed_height), (bed_width, 0), (bed_width, bed_height)],
            "center": [(-bed_width / 2, bed_height / 2), (-bed_width / 2, -bed_height / 2),
                       (bed_width / 2, bed_height / 2),
                       (bed_width / 2, -bed_height / 2)]
        }[origin]
#Pedro END

#Pedro BEGIN Changes to support anchoring the bosty workspace to the bottom of the document
        #for i, (x, y) in enumerate(reference_points_svg):
        for i, (x, y) in enumerate(reference_points_svg.values()):
#Pedro END
            reference_point = etree.Element("{%s}g" % svg_name_space)

            stroke_width = 2
            size = 7

            x_direction = -1 if x > 0 else 1
            plus_sign = etree.Element("{%s}g" % svg_name_space)
            horizontal = etree.Element("{%s}line" % svg_name_space)
            horizontal.set("x1", str(x - x_direction * stroke_width / 2))
            horizontal.set("y1", str(y))
            horizontal.set("x2", str(x + x_direction * size))
            horizontal.set("y2", str(y))
            horizontal.set("style", f"stroke:black;stroke-width:{stroke_width}")
            plus_sign.append(horizontal)

#Pedro BEGIN Changes to support anchoring the botsy workspace to the bottom of the document.
            #y_direction = -1 if y > 0 else 1
            y_direction = 1 if y < self.svg.viewbox_height else -1
#Pedro END
            vertical = etree.Element("{%s}line" % svg_name_space)
            vertical.set("x1", str(x))
            vertical.set("y1", str(y + stroke_width / 2))
            vertical.set("x2", str(x))
            vertical.set("y2", str(y + y_direction * size))
            vertical.set("style", f"stroke:black;stroke-width:{stroke_width}")
            plus_sign.append(vertical)

            reference_point.append(plus_sign)

            text_box = etree.Element("{%s}text" % svg_name_space)
            text_box.set("x", str(x - 28))
#Pedro BEGIN Changes to support anchoring the botsy workspace to the bottom of the document.
            #text_box.set("y", str(y - (y <= 0) * 6 + (y > 0) * 9))
            text_box.set("y", str(y - (y < self.svg.viewbox_height) * 6 + (y >= self.svg.viewbox_height) * 9))
#Pedro END
            text_box.set("font-size", "6")
            text_box.text = f"{reference_points_gcode[i][0]}{unit}, {reference_points_gcode[i][1]}{unit}"
            reference_point.append(text_box)

            group.append(reference_point)

        root.append(group)

    def select_non_debug_layer(self):
        """
        Select content_layer and create one if it doesn't exist. This helps stop the user from accidentally placing new
        objects in debug layers.
        """

        root = self.document.getroot()

        unique_id = "layer89324"
        content_layer = root.find("{%s}g[@id='%s']" % (svg_name_space, unique_id))

        if content_layer is None:
            content_layer = etree.Element("{%s}g" % svg_name_space)
            content_layer.set("id", unique_id)
            content_layer.set("{%s}groupmode" % inkscape_name_space, "layer")
            content_layer.set("{%s}label" % inkscape_name_space, "content layer")

        sodipodi = root.find("{%s}namedview" % sodipodi_name_space)
        if sodipodi is not None:
            sodipodi.set("{%s}current-layer" % inkscape_name_space, unique_id)

        root.append(content_layer)

    def clear_debug(self):
        """Removes debug groups. Used before parsing paths for gcode."""
        root = self.document.getroot()

        debug_traces = root.find("{%s}g[@id='debug_traces']" % svg_name_space)
        debug_references = root.find("{%s}g[@id='debug_references']" % svg_name_space)

        if debug_traces is not None:
            root.remove(debug_traces)

        if debug_references is not None:
            root.remove(debug_references)

    def add_arguments(self, arg_parser):
        """Tell inkscape what arguments to stick in self.options (behind the hood it's more complicated, see docs)"""
        arguments = self.read_arguments()

        for arg in arguments:
            arg_parser.add_argument("--" + arg["name"], type=arg["type"], dest=arg["name"])

    @staticmethod
    def read_arguments():
        """
        This method reads arguments off of the inx file so you don't have to explicitly declare them in self.add_arguments()
        """
        root = etree.parse(inx_filename).getroot()

        arguments = []  # [{name, type, ...}]
        namespace = "http://www.inkscape.org/namespace/inkscape/extension"
        for arg in root.iter("{%s}param" % namespace):

            name = arg.attrib["name"]

            arg_type = arg.attrib["type"]

            if arg_type in ["description", "notebook"]:
                continue

            types = {"int": int, "float": float, "bool": Boolean, "string": str, "optiongroup": str, "path": str}

            arguments.append({"name": name, "type": types[arg_type]})

        if next(root.iter("{%s}page" % namespace)) is not None:
            arguments.append({"name": "tabs", "type": str})

        return arguments


if __name__ == '__main__':
    effect = GcodeExtension()
    effect.run()
